from gensim.models import KeyedVectors
import torch
import torch.autograd as autograd
import re
import SentenceTokenizer as st
import csv


def load_fasttext_embeddings(path, pad):
    fasttext = KeyedVectors.load(path, mmap=None)
    fasttext.add(pad, [0] * fasttext.vector_size, replace=True)
    return fasttext

def load_postag_representation(embeddings, vocab):
    embeddings.weight.data = torch.eye(len(vocab))


class PeriodChecker(object):
    PATTERN = re.compile(r'([0-9]+)(e|i|)')

    def __init__(self, s: str):
        m = self.PATTERN.match(s)
        assert m is not None, \
            f'String must be in the format of [integer]["e"/"i"/"s"].'

        self.scalar = int(m.group(1))
        self.unit = m.group(2)

    def __call__(self, epochs=None, iters=None, steps=None):
        assert epochs or iters or steps, \
            'One of the items must be provided.'

        if self.unit == 'e':
            if epochs is not None:
                return epochs % self.scalar == 0
        elif self.unit == 'i':
            if iters is not None:
                return iters % self.scalar == 0
        elif self.unit == 's':
            if steps is not None:
                return steps % self.scalar == 0
        else:
            raise ValueError(f'Unrecognized unit: {self.unit}')

        return False


def prepare_batch(sentences, labels, lens):  ###igual ao de baselstmcrf
    lens, idx = torch.sort(lens, 0, True)
    if labels is not None:
        sentences, labels = sentences[:, idx], labels[idx]
        labels = autograd.Variable(labels)
    else:
        sentences = sentences[:, idx]

    sentences = autograd.Variable(sentences)
    lens = autograd.Variable(lens)
    return sentences, labels, lens


def tighten(sequences, lens):
    return [sequence[:seq_len] for sequence, seq_len in zip(sequences, lens)]


def lexicalize(sequence, vocab):
    return [vocab.i2f[word] if word in vocab else '<unk>' for word in sequence]


def unlexicalize(sentence, vocab):
    return [vocab.f2i[word] if word in vocab else vocab.f2i['<unk>'] for word in sentence]

def lexicalize_sequence(sequences, lens, vocab):
    return [lexicalize(sequence, vocab) for sequence in tighten(sequences, lens)]

def pre_process(sentence):
    tokenizer = st.SentenceTokenizer()
    return tokenizer.process_message(sentence)

def create_save_file(path, use_lstm_pred=False):
    with open(path, 'w') as file:
        writer = csv.writer(file, delimiter='|', lineterminator='\n')
        writer_parameter = ['Sentence', 'POS Tagging', 'NER', 'LSTM'] if use_lstm_pred else ['Sentence', 'POS Tagging', 'NER']
        writer.writerow(writer_parameter)

def save_predict(path, sequence, pred, lstm=None):
    with open(path, 'a') as file:
        writer = csv.writer(file, delimiter='|', lineterminator='\n')
        if lstm is not None:
            writer_parameter = zip([' '.join(k) for k in sequence[0]], [' '.join(k) for k in sequence[1]],
                                   [' '.join(k) for k in pred], [' '.join(k) for k in lstm])
        else:
            writer_parameter = zip([' '.join(k) for k in sequence[0]], [' '.join(k) for k in sequence[1]],
                                   [' '.join(k) for k in pred])
        for row in writer_parameter:
            writer.writerow(row)


import os
import logging

import numpy as np
import torch
import torch.optim as optim

from TakeBlipNer import utils, data, nermodel

import time


class LSTMCRFTrainer(object):
    def __init__(self, bilstmcrf_model: nermodel.LSTMCRF_NER, input_vocab, input_path,
                 label_vocab, postag_model, postag_label_vocab, save_dir, tensorboard,
                 ckpt_period, val, val_period, samples, pad_string, unk_string,
                 batch_size, shuffle, label_column, encoding, separator,
                 use_pre_processing, learning_rate, learning_rate_decay,
                 max_patience, max_decay_num, patience_threshold, val_path=None,
                 epochs=5, optimizer=optim.Adam, n_iter=5, azureml_logger=None):

        self.model = bilstmcrf_model
        self.epochs = epochs
        self.optimizer_cls = optimizer
        self.optimizer = optimizer(self.model.parameters(), learning_rate)
        self.save_dir = save_dir
        self.tensorboard = tensorboard
        self.input_vocab = input_vocab
        self.postag_label_vocab = postag_label_vocab
        self.postag_model = postag_model
        self.inputs_vocab = [input_vocab] + [postag_label_vocab]
        self.label_vocab = label_vocab
        self.input_path = input_path
        self.val_path = val_path
        self.val = val
        self.writer = None
        self.samples = samples
        self.encoding = encoding
        self.label_column = label_column
        self.separator = separator
        self.use_pre_processing = use_pre_processing
        self.pad_string = pad_string
        self.unk_string = unk_string
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.n_iter = n_iter
        self.min_dev_loss = float('inf')
        self.patience = 0
        self.patience_threshold = patience_threshold
        self.max_patience = max_patience
        self.decay_num = 0
        self.max_decay_num = max_decay_num
        self.learning_rate_decay = learning_rate_decay
        self.logic_break = False
        self.azureml_logger = azureml_logger
        self.validation_number = 0

        self.repeatables = {}

        if self.val:
            self.repeatables[val_period] = self.validate

    @staticmethod
    def get_longest_word(vocab):
        lens = [(len(w), w) for w in vocab.f2i]
        return max(lens, key=lambda x: x[0])[1]

    def display_row(self, items, widths):
        assert len(items) == len(widths)

        padded = ['{{:>{}s}}'.format(w).format(item)
                  for item, w in zip(items, widths)]
        logging.info(' '.join(padded))

    def load_data(self, train):
        path = self.input_path if train else self.val_path

        dataset = data.MultiSentWordDataset(
            path=path,
            label_column=self.label_column,
            encoding=self.encoding,
            separator=self.separator,
            use_pre_processing=self.use_pre_processing)

        dataloader = data.MultiSentWordDataLoader(
            dataset=dataset,
            vocabs=[self.input_vocab]+[self.postag_label_vocab]+[self.label_vocab],
            pad_string=self.pad_string,
            unk_string=self.unk_string,
            batch_size=self.batch_size,
            shuffle=self.shuffle,
            tensor_lens=True,
            postag_model=self.postag_model)
        return dataloader

    def display_samples(self, inputs, targets, lstms, crfs):

        inputs = [[utils.lexicalize(s, v) for s in input]
                  for input, v in zip(inputs, self.inputs_vocab)]
        targets = [utils.lexicalize(s, self.label_vocab) for s in targets]
        lstms = [utils.lexicalize(s, self.label_vocab) for s in lstms]
        crfs = [utils.lexicalize(s, self.label_vocab) for s in crfs]
        transposed = list(zip(*(inputs + [lstms, crfs, targets])))
        col_names = [f'INPUT{i + 1:02d}' for i in range(len(inputs))] + \
                    ['LSTM', 'CRF', 'TARGETS']
        vocabs = self.inputs_vocab + [self.label_vocab] * 3
        col_widths = [max(len(self.get_longest_word(v)), len(c))
                      for v, c in zip(vocabs, col_names)]

        for i, sample in enumerate(transposed):
            rows = list(zip(*sample))

            logging.info('')
            logging.info(f'SAMPLE #{i + 1}')
            self.display_row(col_names, col_widths)
            for row in rows:
                self.display_row(row, col_widths)

    @staticmethod
    def random_idx(max_count, subset=None):
        idx = np.random.permutation(np.arange(max_count))

        if subset is not None:
            return idx[:subset]

        return idx

    @staticmethod
    def gather(lst, idx):
        return [lst[i] for i in idx]

    def save_model(self, steps=None):
        save_path = os.path.join(self.save_dir, 'nermodel.pkl')
        torch.save(self.model, save_path)
        if self.azureml_logger:
            os.system('cp %s %s' % ('nermodel.py', self.save_dir))
            self.azureml_logger.register_ner_model(
                training_mount_folder=self.save_dir,
                ner_registry='NerModel'
            )

    def on_epoch_complete(self, epoch_idx, global_iter, global_step):
        for period_checker, func in self.repeatables.items():
            if period_checker(epochs=epoch_idx,
                              iters=global_iter,
                              steps=global_step):
                func(steps=global_step)
                self.model.train(True)

    def train(self):
        self.model.train(True)
        global_step = 0
        global_iter = 0

        start = time.time()
        for epoch_idx in range(self.epochs):
            epoch_idx += 1
            train_data = self.load_data(train=True)
            for iteration_idx, (batch, lens, _) in enumerate(train_data):
                batch_size = batch[0].size(0)
                iteration_idx += 1
                global_step += batch_size
                global_iter += 1

                self.model.zero_grad()

                batch_sentences, batch_labels = batch[:-1], batch[-1]
                batch_sentences, batch_labels_var, lens_s = utils.prepare_batch(batch_sentences, batch_labels, lens)
                loglik = self.model.loglik(batch_sentences, batch_labels_var, lens_s)
                negative_loglik = -loglik.mean()
                negative_loglik.backward()
                self.optimizer.step()

                if global_step % (self.batch_size * self.n_iter) == 0:
                    logging.info(f'iteration = {global_iter}  negative_loglik={negative_loglik:.4f}')
                    if self.logic_break:
                        logging.info('Early stopping')
                        logging.info('Saving model ...')
                        self.save_model()
                        break
            train_data.dataset.close_file()
            self.on_epoch_complete(epoch_idx, global_iter, global_step)
            if self.logic_break:
                logging.info('Early stopping')
                logging.info('Saving model ...')
                self.save_model()
                break
        logging.info(f'time: {time.time()-start}')
        logging.info(f'iteration ={global_iter}')
        logging.info(f'negative_loglik={negative_loglik:.4f}')
        logging.info('Saving model ...')
        self.save_model()

    def check_loss(self, loss):
        if loss < self.min_dev_loss * self.patience_threshold:
            self.min_dev_loss = loss
            self.patience = 0
        else:
            self.patience += 1
            if self.patience == self.max_patience:
                self.decay_num += 1
                if self.decay_num == self.max_decay_num:
                    self.logic_break = True
                learning_rate = self.optimizer.param_groups[0]['lr'] * self.learning_rate_decay
                for param_group in self.optimizer.param_groups:
                    param_group['lr'] = learning_rate
                self.patience = 0

    def validate(self, steps=None):
        if not self.val:
            return

        path = os.path.join(self.save_dir, 'predict_validation.csv')
        utils.create_save_file(path)

        logging.info('...............................')
        logging.info('Validating...')
        self.model.train(False)
        data_size = 0
        sampled = False

        val_data = self.load_data(train=False)

        nb_classes = len(self.label_vocab)

        confusion_matrix = torch.zeros(nb_classes, nb_classes)

        loss = 0

        for i_idx, (batch, lens, _) in enumerate(val_data):
            i_idx += 1
            batch_size = batch.size(1)
            data_size += batch_size

            sequence_batch, label_batch = batch[:-1], batch[-1]
            with torch.no_grad():
                sequence_batch_wrapped, label_batch_wrapped, lens_s = utils.prepare_batch(sequence_batch, label_batch, lens)
                loglik, logits = self.model.loglik(sequence_batch_wrapped, label_batch_wrapped, lens_s, return_logits=True)
                preds, _, _ = self.model.predict(sequence_batch_wrapped, lens_s)
            loss += -loglik.sum()
            pred_list = preds.data.tolist()
            targets = label_batch_wrapped.data.tolist()
            lens_s = lens_s.data.tolist()
            sequence_batch = sequence_batch_wrapped.data.tolist()
            preds_all = utils.lexicalize_sequence(pred_list, lens_s, self.label_vocab)
            sequence_batch_all = [
                utils.lexicalize_sequence(sequence, lens_s, vocab) for sequence, vocab in zip(
                    utils.tighten(sequence_batch, lens_s), self.inputs_vocab)]
            utils.save_predict(path, sequence_batch_all, preds_all)

            for t, p in zip(label_batch_wrapped.view(-1), preds.view(-1)):
                confusion_matrix[t.long(), p.long()] += 1

            if not sampled and self.samples > 0:
                sample_idx = self.random_idx(batch_size, self.samples)
                sequence_batch = sequence_batch_wrapped.data.tolist()
                sequence_batch = [utils.tighten(x, lens_s) for x in sequence_batch]
                lstm = logits.max(2)[1]
                lstm = lstm.data.tolist()
                lstm = utils.tighten(lstm, lens_s)

                sequence_batch_smp = [self.gather(x, sample_idx) for x in sequence_batch]
                label_batch_smp = self.gather(targets, sample_idx)
                crf_smp = self.gather(pred_list, sample_idx)
                lstm_smp = self.gather(lstm, sample_idx)

                self.display_samples(sequence_batch_smp, label_batch_smp, lstm_smp, crf_smp)
                del sequence_batch, lstm, sequence_batch_smp, label_batch_smp, crf_smp, lstm_smp

                sampled = True

        val_data.dataset.close_file()
        loss = loss/data_size

        self.check_loss(loss)

        confusion_matrix = confusion_matrix[2:, 2:]
        print(confusion_matrix)
        logging.info(f'precision: {confusion_matrix.diag() / confusion_matrix.sum(dim=0)}')
        logging.info(f'recall: {confusion_matrix.diag() / confusion_matrix.sum(dim=1)}')
        logging.info(f'negative_loglik validate {loss:.4f}')
        if self.azureml_logger: 
            self.azureml_logger.save_confusion_matrix_from_tensor(
                confusion_matrix=confusion_matrix,
                labels=list(self.label_vocab.f2i.keys())[2:],
                current_epoch=self.validation_number
            )
            self.azureml_logger.save_metrics(
                confusion_matrix=confusion_matrix,
                labels=list(self.label_vocab.f2i.keys())[2:]
            )
            self.azureml_logger.save_validation_metrics(loss)
        
        self.validation_number += 1


import torch
import pickle
import logging
from TakeBlipNer import utils
from TakeBlipNer import vocab
from TakeBlipNer import data


class NerPredict:
    def __init__(self, model, label_path, pad_string, unk_string, postag_model,
                 save_dir=None, encoding=None, separator=None):
        self.model = model
        self.label_path = label_path
        self.pad_string = pad_string
        self.unk_string = unk_string
        self.label_vocab = self.read_label_vocab()
        self.postag_model = postag_model
        self.postag_vocab = postag_model.label_vocab
        self.save_dir = save_dir
        self.encoding = encoding
        self.separator = separator
        self.fasttext = postag_model.fasttext

    def create_input_vocab(self, input_sentence):
        vocabulary = vocab.Vocabulary()
        vocabulary.add(self.pad_string)
        vocabulary.add(self.unk_string)
        vocab.populate_vocab(input_sentence, vocabulary)
        return vocabulary

    def read_label_vocab(self):
        with open(self.label_path, 'rb') as f:
            label_vocab = pickle.load(f)
        return label_vocab

    def load_embedding(self, input_vocab):
        self.model.embeddings[0].weight.data = torch.from_numpy(
            self.fasttext[input_vocab.i2f.values()])
        self.model.embeddings[0].weight.requires_grad = False

    def predict_batch(self, filepath, sentence_column, batch_size, shuffle,
                      use_pre_processing, output_lstm, sentences=None):
        self.model.train(False)

        input_vocab = vocab.create_vocabulary(
            input_path=filepath,
            column_name=sentence_column,
            pad_string=self.pad_string,
            unk_string=self.unk_string,
            encoding=self.encoding,
            separator=self.separator,
            use_pre_processing=use_pre_processing,
            sentences=sentences)

        if sentences:
            dataset = data.MultiSentWordBatchDataset(
                use_pre_processing=use_pre_processing,
                sentences=sentences)
            predictions = []
            use_index = True
        else:
            dataset = data.MultiSentWordDataset(
                path=filepath,
                label_column=None,
                encoding=self.encoding,
                separator=self.separator,
                use_pre_processing=use_pre_processing
            )
            utils.create_save_file(self.save_dir, output_lstm)
            use_index = False

        data_load = data.MultiSentWordDataLoader(
            dataset=dataset,
            vocabs=[input_vocab] + [self.postag_vocab],
            pad_string=self.pad_string,
            unk_string=self.unk_string,
            batch_size=batch_size,
            shuffle=shuffle,
            tensor_lens=True,
            use_index=use_index,
            postag_model=self.postag_model)

        logging.info('Updating embedding')

        self.model.device = 'cpu'
        self.model.update_embedding(len(input_vocab))
        self.load_embedding(input_vocab)

        global_step = 0
        logging.info('Embedding updated')

        for i_idx, (batch, lens, index) in enumerate(data_load):
            batch_size = batch[0].size(0)
            global_step += batch_size
            with torch.no_grad():
                sequence_batch_wrapped, index_wrapped, lens_s = utils.prepare_batch(
                    batch, index, lens)
                preds, _, logits = self.model.predict(sequence_batch_wrapped,
                                                      lens_s)
            preds = preds.data.tolist()
            sequence_batch = sequence_batch_wrapped.data.tolist()[0]
            postag_batch = sequence_batch_wrapped.data.tolist()[1]
            preds_all = utils.lexicalize_sequence(preds, lens_s,
                                                  self.label_vocab)
            sequence_batch_all = [
                utils.lexicalize_sequence(sequence, lens_s, vocab) for
                sequence, vocab in
                zip([sequence_batch, postag_batch],
                    [input_vocab, self.postag_vocab])]
            lstm_all = None

            if output_lstm:
                lstm_preds = logits.max(2)[1]
                lstm_preds_list = lstm_preds.data.tolist()
                lstm_all = utils.lexicalize_sequence(lstm_preds_list, lens_s,
                                                     self.label_vocab)

            if sentences:
                index_all = index_wrapped.data.tolist()
                predictions += [{'id': id,
                                 'processed_sentence': ' '.join(sentence),
                                 'postaggings': ' '.join(tag),
                                 'entities': ' '.join(pred)}
                                for id, sentence, tag, pred in
                                zip(index_all, sequence_batch_all[0],
                                    sequence_batch_all[1], preds_all)]
            else:
                utils.save_predict(self.save_dir, sequence_batch_all,
                                   preds_all, lstm_all)

            logging.info(f'iteration ={global_step}')

        if sentences:
            return predictions

    def predict_line(self, line, use_pre_processing=True):
        self.model.train(False)
        if use_pre_processing:
            line = utils.pre_process(line)
        if len(line) == 0:
            return ' ', ' ', ' '
        split_line = line.split()
        input_vocab = self.create_input_vocab([split_line])
        self.model.device = 'cpu'
        self.model.update_embedding(len(input_vocab))
        self.load_embedding(input_vocab)
        processed_line, postag_prediction = self.postag_model.predict_line(
            line, use_pre_processing=False)
        postag_prediction = postag_prediction.split()
        unlexicalized_line = utils.unlexicalize(split_line, input_vocab)
        unlexicalized_postags = utils.unlexicalize(postag_prediction,
                                                   self.postag_vocab)
        with torch.no_grad():
            line_len = torch.LongTensor([len(split_line)])
            sentence = torch.LongTensor([unlexicalized_line]).unsqueeze(0)
            sentence_postags = torch.LongTensor(
                [unlexicalized_postags]).unsqueeze(0)
            collated_line = torch.cat([sentence, sentence_postags])
            predicted_line, _, _ = self.model.predict(collated_line, line_len)
        predicted_str = [' '.join(k) for k in [
            utils.lexicalize(sequence, self.label_vocab)
            for sequence in predicted_line.data.tolist()]]
        postag_tags = ' '.join(postag_prediction)
        return processed_line, postag_tags, predicted_str[0]



import os
import pickle
import csv
import logging
import SentenceTokenizer as st

class Vocabulary(object):
    def __init__(self):
        self.f2i = {}
        self.i2f = {}

    def add(self, word, ignore_duplicates=True):
        if word in self.f2i:
            if not ignore_duplicates:
                raise ValueError(f'"{word}" already exists')
        else:
            idx = len(self.f2i)
            self.f2i[word] = idx
            self.i2f[idx] = word
        return self.f2i[word]

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.i2f[item]
        elif isinstance(item, str):
            return self.f2i[item]
        elif hasattr(item, '__iter__'):
            return [self[ele] for ele in item]
        else:
            raise ValueError(f'Unknown type: "{type(item)}"')

    def __contains__(self, item):
        return item in self.f2i or item in self.i2f

    def __len__(self):
        return len(self.f2i)

def populate_vocab(sentences, vocab):
    for sentence in sentences:
        for word in sentence:
            vocab.add(word)

def create_vocabulary(input_path, column_name, pad_string, unk_string, encoding, separator, use_pre_processing = False, is_label = False, sentences=None):
    vocabulary = Vocabulary()
    vocabulary.add(pad_string)
    vocabulary.add(unk_string)
    if is_label:
        use_pre_processing = False

    if sentences:
        #logging.info('Pre processing...')
        tokenizer = st.SentenceTokenizer()
        sentences = [tokenizer.process_message(sentence['sentence']).split() for sentence in sentences]
    else:
        sentences = read_sentences(input_path, column_name, encoding, separator, use_pre_processing)
    populate_vocab(sentences, vocabulary)
    return vocabulary

def save_vocabs(save_path, input_vocab, file_name):
    vocab_path = os.path.join(save_path, file_name)
    pickle.dump(input_vocab, open(vocab_path, 'wb'))

def read_sentences(path, column, encoding, separator, use_pre_processing):
    with open(path, newline='', encoding=encoding) as file:
        reader = csv.DictReader(file, delimiter = separator)
        if use_pre_processing:
            #logging.info('Pre processing...')
            tokenizer = st.SentenceTokenizer()
            for line in reader:
                yield tokenizer.process_message(line[column]).split()
            #logging.info('Finished pre processing...')
        else:
            for line in reader:
                yield line[column].split()


import torch
import torch.nn as nn
from torch.autograd import Variable

def log_sum_exp(vec, dim=0):
    max, idx = torch.max(vec, dim)
    max_exp = max.unsqueeze(-1).expand_as(vec)
    return max + torch.log(torch.sum(torch.exp(vec - max_exp), dim))

class CRF(nn.Module):
    def __init__(self, vocab_size, pad_idx, unk_idx, device):
        super(CRF, self).__init__()
        self.device = device
        self.vocab_size = vocab_size
        self.n_labels = n_labels = vocab_size + 2
        self.start_idx = n_labels - 2
        self.stop_idx = n_labels - 1
        self.transitions = nn.Parameter(torch.randn(n_labels, n_labels).to(self.device))
        self.pad_idx = pad_idx
        self.unk_idx = unk_idx

    def reset_parameters(self):
        nn.init.constant_(self.transitions.data, 0)
        nn.init.constant_(self.transitions.data[:, self.unk_idx], -3)
        nn.init.constant_(self.transitions.data[:, self.pad_idx], -3)
        nn.init.constant_(self.transitions.data[:, self.start_idx], -3)
        nn.init.constant_(self.transitions.data[:, self.stop_idx], -3)


    def forward(self, logits, lens):
        '''
        Arguments:
            logits: [batch_size, seq_len, n_labels] FloatTensor
            lens: [batch_size] LongTensor
        '''
        batch_size, seq_len, n_labels = logits.size()
        alpha = logits.data.new(batch_size, self.n_labels).fill_(-10000)
        alpha[:, self.start_idx] = 0
        alpha = Variable(alpha)
        cloned_lens = lens.clone()

        logits_transposed = logits.transpose(1, 0)
        for logit in logits_transposed:
            logit_expanded = logit.unsqueeze(-1).expand(batch_size,
                                                        *self.transitions.size())
            alpha_expanded = alpha.unsqueeze(1).expand(batch_size,
                                                       *self.transitions.size())
            transition_expanded = self.transitions.unsqueeze(0).expand_as(alpha_expanded)
            matrix = transition_expanded + alpha_expanded + logit_expanded
            alpha_next = log_sum_exp(matrix, 2).squeeze(-1)

            mask = (cloned_lens > 0).float().unsqueeze(-1).expand_as(alpha)
            alpha = mask * alpha_next + (1 - mask) * alpha
            cloned_lens = cloned_lens - 1

        alpha = alpha + self.transitions[self.stop_idx].unsqueeze(0).expand_as(alpha)
        norm = log_sum_exp(alpha, 1).squeeze(-1)

        return norm

    def viterbi_decode(self, logits, lens):
        '''Borrowed from pytorch tutorial

        Arguments:
            logits: [batch_size, seq_len, n_labels] FloatTensor
            lens: [batch_size] LongTensor
        '''
        batch_size, seq_len, n_labels = logits.size()
        viterbi = logits.data.new(batch_size, self.n_labels).fill_(-10000)
        viterbi[:, self.start_idx] = 0
        viterbi = Variable(viterbi)
        cloned_lens = lens.clone()

        logits_transposed = logits.transpose(1, 0)
        pointers = []
        for logit in logits_transposed:
            viterbi_expanded = viterbi.unsqueeze(1).expand(batch_size, n_labels, n_labels)
            transition_expanded = self.transitions.unsqueeze(0).expand_as(viterbi_expanded)
            viterbi_transition_sum = viterbi_expanded + transition_expanded
            viterbi_max, viterbi_argmax = viterbi_transition_sum.max(2)

            viterbi_max = viterbi_max.squeeze(-1)
            viterbi_next = viterbi_max + logit
            pointers.append(viterbi_argmax.squeeze(-1).unsqueeze(0))

            mask = (cloned_lens > 0).float().unsqueeze(-1).expand_as(viterbi_next)
            viterbi = mask * viterbi_next + (1 - mask) * viterbi

            mask = (cloned_lens == 1).float().unsqueeze(-1).expand_as(viterbi_next)
            viterbi += mask * self.transitions[self.stop_idx].unsqueeze(0).expand_as(viterbi_next)

            cloned_lens = cloned_lens - 1

        pointers = torch.cat(pointers)
        scores, idx = viterbi.max(1)
        paths = [idx.unsqueeze(1)]

        for argmax in reversed(pointers):
            idx_exp = idx.unsqueeze(-1)
            idx = torch.gather(argmax, 1, idx_exp)
            idx = idx.squeeze(-1)

            paths.insert(0, idx.unsqueeze(1))

        paths = torch.cat(paths[1:], 1)
        scores = scores.squeeze(-1)

        return scores, paths

    def transition_score(self, labels, lens):
        '''
        Arguments:
             labels: [batch_size, seq_len] LongTensor
             lens: [batch_size] LongTensor
        '''
        batch_size, seq_len = labels.size()

        # pad labels with <start> and <stop> indices
        labels_ext = Variable(labels.data.new(batch_size, seq_len + 2))
        labels_ext[:, 0] = self.start_idx
        labels_ext[:, 1:-1] = labels
        mask = sequence_mask(lens + 1, self.device, max_len=seq_len + 2).long()
        pad_stop = Variable(labels.data.new(1).fill_(self.stop_idx))
        pad_stop = pad_stop.unsqueeze(-1).expand(batch_size, seq_len + 2)
        labels_ext = (1 - mask) * pad_stop + mask * labels_ext
        labels = labels_ext

        transitions = self.transitions

        # obtain transition vector for each label in batch and timestep
        # (except the last ones)
        transitions_expanded = transitions.unsqueeze(0).expand(batch_size, *transitions.size())
        labels_except_last = labels[:, 1:]
        labels_except_last_expanded = labels_except_last.unsqueeze(-1)\
            .expand(*labels_except_last.size(), transitions.size(0))
        transitions_row = torch.gather(transitions_expanded, 1, labels_except_last_expanded)

        # obtain transition score from the transition vector for each label
        # in batch and timestep (except the first ones)
        labels_except_first_expanded = labels[:, :-1].unsqueeze(-1)
        transitions_score = torch.gather(transitions_row, 2, labels_except_first_expanded)
        transitions_score = transitions_score.squeeze(-1)

        mask = sequence_mask(lens + 1, self.device).float()
        transitions_score = transitions_score * mask
        score = transitions_score.sum(1).squeeze(-1)
        return score


class LSTMCRF_NER(nn.Module):
    def __init__(self, crf, vocab_sizes, word_dims, hidden_dim, layers,
                 dropout_prob, device, alpha=0, bidirectional=False):
        super(LSTMCRF_NER, self).__init__()

        self.total_word_dim = sum(word_dims)
        self.n_feats = len(word_dims)
        self.device = device
        self.word_dims = word_dims
        self.hidden_dim = hidden_dim
        self.lstm_layers = layers
        self.dropout_prob = dropout_prob
        self.alpha = alpha

        self.crf = crf
        self.bidirectional = bidirectional
        self.n_labels = n_labels = self.crf.n_labels
        self.embeddings = nn.ModuleList([nn.Embedding(vocab_size, word_dim).to(self.device)
                                         for vocab_size, word_dim in zip(vocab_sizes, word_dims)]).to(self.device)

        self.output_hidden_dim = self.hidden_dim
        if bidirectional:
            self.output_hidden_dim *= 2

        self.tanh = nn.Tanh()
        self.input_layer = nn.Linear(self.total_word_dim, hidden_dim)
        self.output_layer = nn.Linear(self.output_hidden_dim, n_labels)
        self.lstm = nn.LSTM(input_size=hidden_dim,
                            hidden_size=hidden_dim,
                            num_layers=layers,
                            bidirectional=bidirectional,
                            dropout=dropout_prob,
                            batch_first=True)

    def reset_parameters(self):
        for emb in self.embeddings:
            nn.init.xavier_normal_(emb.weight.data)

        nn.init.xavier_normal_(self.input_layer.weight.data)
        nn.init.xavier_normal_(self.output_layer.weight.data)
        self.crf.reset_parameters()
        self.lstm.reset_parameters()

    def update_embedding(self, vocab_size):
        self.embeddings[0] = nn.Embedding(vocab_size, self.word_dims[0]).to(self.device)

    def _run_rnn_packed(self, cell, sequence_batch, sequence_batch_lens, h=None):
        sequence_batch_packed = nn.utils.rnn.pack_padded_sequence(sequence_batch, sequence_batch_lens.data.tolist(),
                                                  batch_first=True)

        if h is not None:
            output, h = cell(sequence_batch_packed, h)
        else:
            output, h = cell(sequence_batch_packed)

        output, _ = nn.utils.rnn.pad_packed_sequence(output, batch_first=True)

        return output, h

    def _embeddings(self, sequences_batch):
        '''Takes raw feature sequences and produces a single word embedding

        Arguments:
            xs: [n_feats, batch_size, seq_len] LongTensor

        Returns:
            [batch_size, seq_len, word_dim] FloatTensor 
        '''

        n_feats, batch_size, seq_len = sequences_batch.size()
        assert n_feats == self.n_feats

        embedded_sequence_combination = [emb(sequence) for emb, sequence in zip(self.embeddings.to(self.device), sequences_batch.to(self.device))]
        embedded_sequence_combination_cat = torch.cat(embedded_sequence_combination, 2)

        return embedded_sequence_combination_cat

    def _forward_bilstm(self, sequences_batch, lens):
        n_feats, batch_size, seq_len = sequences_batch.size()

        embedded_sequences_batch = self._embeddings(sequences_batch)
        embedded_sequences_batch = embedded_sequences_batch.view(-1, self.total_word_dim)
        embedded_sequences_batch = self.tanh(self.input_layer(embedded_sequences_batch))
        embedded_sequences_batch = embedded_sequences_batch.view(batch_size, seq_len, self.hidden_dim)

        output, h = self._run_rnn_packed(self.lstm, embedded_sequences_batch, lens)

        output = output.contiguous()
        output = output.view(-1, self.output_hidden_dim)
        output = self.tanh(self.output_layer(output))
        output = output.view(batch_size, seq_len, self.n_labels)

        return output

    def _bilstm_score(self, logits, sequence_labels, lens):
        sequence_labels_expanded = sequence_labels.unsqueeze(-1)
        scores = torch.gather(logits, 2, sequence_labels_expanded).squeeze(-1)
        mask = sequence_mask(lens, self.device).float()
        scores = scores * mask
        score = scores.sum(1).squeeze(-1)

        return score

    def score(self, sequences, sequence_labels, lens, logits=None):
        if logits is None:
            logits = self._forward_bilstm(sequences, lens)

        transition_score = self.crf.transition_score(sequence_labels, lens)
        bilstm_score = self._bilstm_score(logits, sequence_labels, lens)

        score = transition_score + bilstm_score

        return score

    def predict(self, sequences, lens):
        logits = self._forward_bilstm(sequences, lens)
        scores, preds = self.crf.viterbi_decode(logits, lens)

        return preds, scores, logits

    def loglik(self, sequences, sequence_labels, lens, return_logits=False):
        logits = self._forward_bilstm(sequences, lens)
        norm_score = self.crf(logits, lens)
        sequence_score = self.score(sequences, sequence_labels, lens, logits=logits)
        loglik = sequence_score - norm_score - self.alpha*self.crf.transitions[2:-2, 2:-2].pow(2).sum()

        if return_logits:
            return loglik, logits
        else:
            return loglik

def sequence_mask(lens, device, max_len=None):
    batch_size = lens.size(0)

    if max_len is None:
        max_len = lens.max().data

    ranges = torch.arange(0, max_len).long().to(device)
    ranges = ranges.unsqueeze(0).expand(batch_size, max_len)
    ranges = Variable(ranges)

    lens_expanded = lens.unsqueeze(1).expand_as(ranges)
    mask = ranges < lens_expanded

    return mask


import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.metrics import confusion_matrix
from azureml.core import Run, Model

class AzuremlLogger(object):
    def __init__(self, run, torch_version):
        self.run = run
        self.torch_version = torch_version

    def save_train_parameters(self, args):
        self.run.log('input-path', args.input_path)
        self.run.log('separator', args.separator)
        self.run.log('encoding', args.encoding)
        self.run.log('sentence_column', args.sentence_column)
        self.run.log('label_column', args.label_column)
        self.run.log('save-dir', args.save_dir)
        self.run.log('tensorboard', args.tensorboard)
        self.run.log('use_pre_processing', args.use_pre_processing)
        self.run.log('wordembed-path', args.wordembed_path)
        self.run.log('postag_model_path', args.postag_model_path)
        self.run.log('postag_label_path', args.postag_label_path)
        self.run.log('epochs', args.epochs)
        self.run.log('dropout-prob', args.dropout_prob)
        self.run.log('batch-size', args.batch_size)
        self.run.log('shuffle', args.shuffle)
        self.run.log('learning-rate', args.learning_rate)
        self.run.log('learning-rate-decay', args.learning_rate_decay)
        self.run.log('max-patience', args.max_patience)
        self.run.log('max-decay-num', args.max_decay_num)
        self.run.log('patience-threshold', args.patience_threshold)
        self.run.log('ckpt-period', args.ckpt_period)
        self.run.log('val', args.val)
        self.run.log('val-path', args.val_path)
        self.run.log('val-period', args.val_period)
        self.run.log('samples', args.samples)
        self.run.log('word-dim', args.word_dim)
        self.run.log('lstm-dim', args.lstm_dim)
        self.run.log('lstm-layers', args.lstm_layers)
        self.run.log('bidirectional', args.bidirectional)
        self.run.log('input-data-ref', args.input_data_ref)
        self.run.log('wordembed-model-name', args.wordembed_model_name)
        self.run.log('wordembed-model-version', args.wordembed_model_version)
        self.run.log('alpha', args.alpha)

    def save_validation_metrics(self, metric):
        self.run.log('negative_loglik', metric.item())
        
    def save_confusion_matrix(self, targets_all, preds_all, current_epoch):
        image_file_name = 'confusion_matrix_epoch_{}.png'.format(current_epoch)
        labels = list(set(targets_all))
        labels.sort()
        cm = confusion_matrix(targets_all, preds_all)
        plt.figure(figsize=(16,10))
        sns.heatmap(cm, annot=True, cmap=plt.cm.Blues, xticklabels=labels, yticklabels=labels,  fmt='d')
        plt.yticks(rotation=0) 
        plt.savefig(image_file_name)
        self.run.log_image(image_file_name, plot=plt)

    def save_confusion_matrix_from_tensor(self, confusion_matrix, labels, current_epoch):
        image_file_name = 'confusion_matrix_validation_{}.png'.format(current_epoch)
        plt.figure(figsize=(16,10))
        sns.heatmap(confusion_matrix.long().numpy(), annot=True, cmap=plt.cm.Blues, xticklabels=labels, yticklabels=labels, fmt='d')
        plt.yticks(rotation=0) 
        plt.savefig(image_file_name)
        self.run.log_image(image_file_name, plot=plt)

    def save_loss_convergence(self, loss):
        image_file_name = 'validation_loss.png'
        fig = plt.figure()
        plt.plot(loss, color='blue')
        plt.legend(['Validation Loss'], loc='upper right')
        plt.xlabel('Number of validating examples')
        plt.ylabel('Negative log likelihood loss')
        fig.savefig(image_file_name)
        self.run.log_image(image_file_name, plot=fig)

    def save_report(self, report):
        self.run.log('Accuracy', report['accuracy'])
        self.run.log('Precision - Macro Avg', report['macro avg']['precision'])
        self.run.log('Recall - Macro Avg', report['macro avg']['recall'])
        self.run.log('F1-score - Macro Avg', report['macro avg']['f1-score'])
        self.run.log('Precision - Weighted Avg', report['weighted avg']['precision'])
        self.run.log('Recall - Weighted Avg', report['weighted avg']['recall'])
        self.run.log('F1-score - Weighted Avg', report['weighted avg']['f1-score'])
    
    def save_metrics(self, confusion_matrix, labels):
        precision = confusion_matrix.diag() / confusion_matrix.sum(dim=0)
        recall = confusion_matrix.diag() / confusion_matrix.sum(dim=1)
        f1_score = 2*(precision*recall / (precision + recall))
        #self.run.log('Accuracy', report['accuracy'])
        for index, label in enumerate(labels):
            self.run.log(label + ' Precision', precision[index].numpy().item())
            self.run.log(label + ' Recall', recall[index].numpy().item())
            self.run.log(label + ' F1-score', f1_score[index].numpy().item())

        self.run.log('Model Precision', precision[precision >= 0].mean().numpy().item())
        self.run.log('Model Recall', recall[recall >= 0].mean().numpy().item())
        self.run.log('Model F1-score', f1_score[f1_score >= 0].mean().numpy().item())

    def register_ner_model(self, training_mount_folder, ner_registry):
        ner_model = Model.register(
            workspace=self.run.experiment.workspace,
            model_name=ner_registry,
            model_path=training_mount_folder,
            child_paths=['nermodel.pkl', 'nermodel.py'],
            description='NER Pytorch Model',
            model_framework=Model.Framework.PYTORCH,  
            model_framework_version=self.torch_version, 
            tags = {'runId': self.run.id}
        )
        self.run.log(ner_registry, ner_model)

    def register_ner_label(self, training_mount_folder, ner_registry_label):
        ner_label= Model.register(
            workspace=self.run.experiment.workspace,
            model_name=ner_registry_label,
            model_path=training_mount_folder,
            child_paths=['ner-vocab-label.pkl', 'vocab.py'],
            description='NER Pytorch Labels',
            tags={'runId': self.run.id}
        )
        self.run.log(ner_registry_label, ner_label)


import csv
import torch
import sys
import SentenceTokenizer as st
from torch.utils.data import DataLoader, IterableDataset, Dataset
import pickle

class MultiSentWordDataset(IterableDataset):
    def __init__(self, path, label_column, encoding, separator, use_pre_processing=True):
        self.use_pre_processing = use_pre_processing
        self.__set_file_iterable(path, encoding, separator)
        self.__set_default_process()
        self.line_mapper = self.line_mapper_training if label_column else \
            self.line_mapper_predicting


    def __set_file_iterable(self, path, encoding, separator):
        self.file_pointer = open(path, encoding=encoding)
        self.file_iterable = csv.reader(self.file_pointer,  delimiter=separator)
        next(self.file_iterable)

    def __set_default_process(self):
        if self.use_pre_processing:
            self.tokenizer = st.SentenceTokenizer()
            self.process = self.pre_process
        else:
            self.process = self.split_sentence

    def close_file(self):
        self.file_pointer.close()

    def line_mapper_training(self, line):
        return [self.process(line[0]), self.split_sentence(line[1])]

    def line_mapper_predicting(self, line):
        return [self.process(line[0])]

    def pre_process(self, sentence):
        return self.tokenizer.process_message(sentence).split()

    def split_sentence(self, sentence):
        return sentence.split()

    def __iter__(self):
        file_iter = self.file_iterable
        return map(self.line_mapper, file_iter)

    def __len__(self):
        return 1

class MultiSentWordBatchDataset(Dataset):
    def __init__(self, use_pre_processing, sentences):
        self.use_pre_processing = use_pre_processing
        self.sentences = sentences
        self.tokenizer = st.SentenceTokenizer()
        self.__set_default_process()

    def pre_process(self, sentence):
        return self.tokenizer.process_message(sentence).split()

    def split_sentence(self, sentence):
        return sentence.split()

    def __set_default_process(self):
        if self.use_pre_processing:
            self.tokenizer = st.SentenceTokenizer()
            self.process = self.pre_process
        else:
            self.process = self.split_sentence

    def __len__(self):
        return len(self.sentences)

    def __getitem__(self, idx):
        return [self.process(self.sentences[idx]['sentence']), int(self.sentences[idx]['id'])]

class MultiSentWordDataLoader(DataLoader):
    def __init__(self, dataset, vocabs, pad_string, unk_string,
                 postag_model=None, input_vocab=None,
                 tensor_lens=True, use_index=False, **kwargs):
        super(MultiSentWordDataLoader, self).__init__(
            dataset=dataset,
            collate_fn=self.collate_fn,
            **kwargs
        )

        self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        self.pad_string = pad_string
        self.unk_string = unk_string
        self.tensor_lens = tensor_lens
        self.vocabs = vocabs
        self.use_index = use_index
        self.postag = postag_model
        self.use_pre_processing = dataset.use_pre_processing

    def unlexicalize(self, sentence):
        return [
            self.sentences_vocab.f2i[word] if word in self.sentences_vocab else
            self.sentences_vocab.f2i[self.unk_string] for word in sentence]

    def pad(self, sentence):
        return sentence + [self.sentences_vocab.f2i[self.pad_string]] * (
                self.max_len - len(sentence))

    def predict_tags_batch(self, batches):
        dict_list = []
        msg_list = []
        ner_list = []
        for idx, row in enumerate(batches):
            if len(row) > 1:
                ner_list.append(row[1])
            dict_list.append({'id': idx, 'sentence': ' '.join(row[0])})
            msg_list.append(row[0])
        tags = sorted(self.postag.predict_batch(filepath='',
                                                sentence_column='',
                                                pad_string='<pad>',
                                                unk_string='<unk>',
                                                batch_size=len(dict_list),
                                                shuffle=True,
                                                use_pre_processing=False,
                                                output_lstm=False,
                                                sentences=dict_list),
                      key=lambda content_dict: content_dict['id'])
        if len(ner_list) > 0:
            batches = [[msg, tag_dict['tags'].split(), ner]
                       for msg, tag_dict, ner in zip(msg_list, tags, ner_list)]
        else:
            batches = [[msg, tag_dict['tags'].split()]
                       for msg, tag_dict in zip(msg_list, tags)]
        return batches


    def collate_fn(self, batches):
        batches = filter(lambda x: len(x[0]) > 0, batches)
        if self.postag:
            batches = self.predict_tags_batch(batches)
        sentences_list = list(zip(*batches))
        lens = [len(sentence) for sentence in sentences_list[0]]
        self.max_len = max(lens)
        sequences_collation = []
        for sentences, sentences_vocab in zip(sentences_list, self.vocabs):
            self.sentences_vocab = sentences_vocab
            sentences = map(self.pad, map(self.unlexicalize, sentences))
            sentences = list(sentences)
            sequences_collation.append(torch.LongTensor(sentences).unsqueeze(0))
        sequence_tensor = torch.cat(sequences_collation).to(self.device)
        lens_tensor = torch.LongTensor(lens).to(self.device)
        if self.use_index:
            index_tensor = torch.LongTensor(sentences_list[-1]).to(self.device)
            return sequence_tensor, lens_tensor, index_tensor
        else:
            return sequence_tensor, lens_tensor, None 

maxInt = sys.maxsize
while True:
    # decrease the maxInt value by factor 10 
    # as long as the OverflowError occurs.
    try:
        csv.field_size_limit(maxInt)
        break
    except OverflowError:
        maxInt = int(maxInt/10)



import setuptools

with open("README_PACKAGE.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="TakeBlipNer",
    version="0.0.4",
    author="Data and Analytics Research",
    author_email="analytics.dar@take.net",
    credits=['Renan Santos'],
	keywords='ner',
    description="Named Entity Recognition Package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
	install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
	include_package_data = True
)